try:
    from IPython import embed # pylint: disable=F0401
except ImportError:
    import code
    def _interact(ns):
        code.interact(local=ns)
else:
    def _interact(ns):
        embed(user_ns=ns)

def start_interactive_shell(**namespace):
    """
    Starts an interactive shell. Uses IPython if available, else fall back
    to the native Python interpreter.

    Any keyword argument specified will be available in the shell ``globals``.
    """
    _interact(namespace)



from .utils import TestCase
from .utils import run_tests_assert_success
from .utils.test_generator import TestGenerator
import itertools
import os
import shutil
import slash
import tempfile

class TestMetadataTest(TestCase):

    def setUp(self):
        super(TestMetadataTest, self).setUp()
        self.root = self.get_new_path()
        self.filename = os.path.join(self.root, "testfile.py")
        with open(self.filename, "w") as f:
            f.write(_TEST_FILE_TEMPLATE)

        self.tests = list((slash.loader.Loader().iter_path(self.filename)))
        self.session = run_tests_assert_success(self.tests)
        self.results = list(self.session.results.iter_test_results())
        self.results.sort(key = lambda result: str(result.test_metadata.fqn))

    def test_tests_have_correct_metadata(self):
        for test, result in zip(self.tests, self.session.results.iter_test_results()):
            self.assertIs(test.__slash__, result.test_metadata)

    def test_simple_test_fqn(self):
        simple_test_fqn = self.results[0].test_metadata.fqn
        self.assertEquals(simple_test_fqn, "{0}:T001.test_method".format(self.filename))

    def test_parameterized_test_fqn(self):
        parameterized = set(str(x.test_metadata.fqn) for x in self.results[1:])

        self.assertEquals(parameterized, set(
            "{0}:T002(a={1})(c={2}).test_parameters(b={3})".format(self.filename, a, c, b)
            for a, b, c in itertools.product([1, 2], [3, 4], [5, 6])))

_TEST_FILE_TEMPLATE = """
import slash

class T001(slash.Test):
    def test_method(self):
        pass

class T002(slash.Test):

    @slash.parameters.iterate(a=[1, 2])
    def before(self, a):
        pass

    @slash.parameters.iterate(b=[3, 4])
    def test_parameters(self, b):
        pass

    @slash.parameters.iterate(c=[5, 6])
    def after(self, c):
        pass
"""



from .utils import TestCase
import slash
import slash.runner
from slash import exception_handling
from slash import Session
from slash.loader import Loader

class CleanupsTest(TestCase):
    def test_cleanups(self):
        class Test(slash.Test):
            def test1(self_):
                self.events.test1()
                slash.add_cleanup(self.events.cleanup, "test1 cleanup 2")
                slash.add_cleanup(self.events.cleanup, "test1 cleanup 1")
            def test2(self_):
                self.events.test2()
                slash.add_cleanup(self.events.cleanup, "test2 cleanup")

        with self.forge.any_order():
            with self.forge.ordered():
                self.events.test1()
                self.events.cleanup("test1 cleanup 1")
                self.events.cleanup("test1 cleanup 2")
            with self.forge.ordered():
                self.events.test2()
                self.events.cleanup("test2 cleanup")

        self.forge.replay()
        with Session():
            slash.runner.run_tests(Loader().iter_test_factory(Test))

    def test_error_cleanups(self):

        exc_infos = []
        self.forge.replace_with(exception_handling, "_EXCEPTION_HANDLERS", [exc_infos.append])
        self.override_config("debug.enabled", True)

        class Test(slash.Test):
            def test(self_):
                slash.add_cleanup(self.events.cleanup, 1)
                slash.add_cleanup(self.events.cleanup, 2)
                raise Exception("!!!")
        self.events.cleanup(2).and_raise(SecondException())
        self.events.cleanup(1).and_raise(FirstException())
        self.forge.replay()
        with Session() as session:
            slash.runner.run_tests(Loader().iter_test_factory(Test))
        [result] = session.results.iter_test_results()
        [err1, err2, err3] = errors = result.get_errors()

        self.assertEquals(len(errors), len(exc_infos))
        self.assertEquals(
            [e[0] for e in exc_infos],
            [Exception, SecondException, FirstException],
            )

class FirstException(Exception):
    pass
class SecondException(Exception):
    pass
class ThirdException(Exception):
    pass



from .metadata import Metadata

class RunnableTestFactory(object):

    @classmethod
    def generate_tests(cls):
        """
        Generates :class:`.RunnableTest` instances to run

        Do not override this method directly. Use :func:`.RunnableTestFactory._generate_tests` instead.
        """
        for index, test in enumerate(cls._generate_tests()):
            assert test.__slash__ is None
            test.__slash__ = Metadata(test, factory=cls, factory_index=index)
            yield test

    @classmethod
    def _generate_tests(cls):
        raise NotImplementedError() # pragma: no cover

    __slash_needed_contexts__ = None




from contextlib import contextmanager
import sys
from . import hooks as trigger_hook
from .loader import Loader
from .core.session import Session
from . import site
from .utils import cli_utils
from .utils.interactive import start_interactive_shell
from .utils.reporter import report_context

class Application(object):
    def __init__(self, parser, args, report_stream):
        super(Application, self).__init__()
        self.parser = parser
        self.args = args
        self.report_stream = report_stream
        self.test_loader = Loader()
        self.session = Session()
    def error(self, *args, **kwargs):
        self.parser.error(*args, **kwargs)

@contextmanager
def get_application_context(parser=None, argv=None, args=(), report_stream=sys.stderr, enable_interactive=False, allow_unknown_args=False):
    site.load()
    args = list(args)
    if enable_interactive:
        args.append(
            cli_utils.Argument("-i", "--interactive", help="Enter an interactive shell",
                               action="store_true", default=False)
        )
    with cli_utils.get_cli_environment_context(argv=argv, extra_args=args, allow_unknown_args=allow_unknown_args) as (parser, parsed_args):
        app = Application(parser=parser, args=parsed_args, report_stream=report_stream)
        with app.session:
            with report_context(app.report_stream):
                if enable_interactive and parsed_args.interactive:
                    start_interactive_shell()
                yield app
            trigger_hook.result_summary()




from .utils import TestCase
import slash
from slash import Session
from slash import context

class GlobalsTest(TestCase):

    def setUp(self):
        super(GlobalsTest, self).setUp()
        self.session = Session()

    def test_get_current_session(self):
        with self.session:
            self.assertIs(context.session, self.session)
            self.assertIsNot(context.session, slash.session)
            self.assertEquals(self.session, slash.session)

    def test_get_current_test(self):
        with self.session:
            self.assertIsNone(context.test)
            self.assertIsNone(context.test_id)
            parent_test = self
            class InnerTest(slash.Test):
                def test_method(self):
                    parent_test.assertIs(context.test, self)
                    parent_test.assertEquals(context.test_id, self.__slash__.id)
                    parent_test.assertEquals(context.test_filename, _without_pyc(__file__))
                    parent_test.assertEquals(context.test_classname, "InnerTest")
                    parent_test.assertEquals(context.test_methodname, "test_method")
            slash.runner.run_tests(InnerTest.generate_tests())
        self.assertTrue(self.session.results.is_success())

def _without_pyc(path):
    if path.endswith(".pyc"):
        path = path[:-1]
    return path



import functools
import sys
from contextlib import contextmanager

import logbook

from ..app import get_application_context
from ..conf import config
from ..exceptions import SlashException
from ..runner import run_tests
from ..utils import rerunning_utils

_logger = logbook.Logger(__name__)

def slash_run(args, report_stream=None, rerun=False):
    if report_stream is None:
        report_stream = sys.stderr
    with _get_slash_app_context(args, report_stream, rerun) as app:
        if rerun:
            try:
                app.prev_session_state = rerunning_utils.get_last_session_state()
            except IOError as e:
                logbook.error("Cannot load state file! ({0})", e)
                return -1
        else:
            app.prev_session_state = None

        iterator = _get_test_iterator(app, args)
        try:
            run_tests(iterator)
        except SlashException as e:
            logbook.error(e)
            return -1
        finally:
            _save_rerun_state(app)

        if app.session.results.is_success(allow_skips=True):
            return 0
        return -1

@contextmanager
def _get_slash_app_context(args, report_stream, rerun):
    with get_application_context(
            argv=args, allow_unknown_args=not rerun,
            enable_interactive=True,
            report_stream=report_stream) as app:
        yield app

slash_rerun = functools.partial(slash_run, rerun=True)

def _get_test_iterator(app, args): # pylint: disable=unused-argument
    if app.prev_session_state:
        return _get_rerun_test_iterator(app)

    paths = app.args.remainder
    if not paths:
        paths = config.root.run.default_sources
    if not paths and not app.args.interactive:
        app.error("No tests specified")

    return app.test_loader.iter_pqns(paths)

def _get_rerun_test_iterator(app):
    saved_results = app.prev_session_state["results"]

    for test in app.test_loader.iter_pqns(app.prev_session_state["pqns"]):
        _logger.debug("Rerun: examining {0}", test)

        saved_result = saved_results.get(test.__slash__.fqn.fqn, None)
        if saved_result is not None and not saved_result["rerun_needed"]:
            _logger.debug("Skipping {0} (no need to rerun)", test)
            continue
        _logger.debug("Rerun: retrying {0}", test)
        yield test

def _save_rerun_state(app):
    if app.prev_session_state is not None:
        state = app.prev_session_state
    else:
        state = {"pqns": app.args.remainder, "results": {}}
    saved_results = state["results"]

    for result in app.session.results.iter_test_results():
        saved_result = saved_results.get(result.test_metadata.fqn.fqn, None)
        if saved_result is None:
            saved_result = saved_results[result.test_metadata.fqn.fqn] = {"rerun_needed": False}
        saved_result["rerun_needed"] = result.is_error() or result.is_failure()

    rerunning_utils.save_session_state(state)



import sys
import traceback
from .._compat import string_types

class Error(object):

    def __init__(self, exc_info_or_msg):
        super(Error, self).__init__()
        if not isinstance(exc_info_or_msg, tuple):
            exc_info = sys.exc_info()
            if isinstance(exc_info_or_msg, string_types):
                message = exc_info_or_msg
            else:
                message = None
        else:
            exc_info = exc_info_or_msg
            message = None
        self.arg = exc_info_or_msg
        self.exception_type, self.exception, tb = exc_info
        self.message = message
        if tb is not None:
            self.exception_text = "".join(traceback.format_exception(
                self.exception_type, self.exception, tb
            ))
        else:
            self.exception_text = None

    def __repr__(self):
        return repr(self.exception)

    def __str__(self):
        return str(self.exception)



from .utils.callback import Callback
from .exceptions import HookAlreadyExists
from . import _compat

session_start = Callback(doc="Called right after session starts")
session_end   = Callback(doc="Called right before the session ends, regardless of the reason for termination")

test_interrupt = Callback(doc="Called when a test is interrupted by a KeyboardInterrupt or other similar means")
test_start   = Callback(doc="Called right after a test starts")
test_end     = Callback(doc="Called right before a test ends, regardless of the reason for termination")
test_success = Callback(doc="Called on test success")
test_error   = Callback(doc="Called on test error")
test_failure = Callback(doc="Called on test failure")
test_skip    = Callback(doc="Called on test skip", arg_names=("reason",))

result_summary = Callback(doc="Called at the end of the execution, when printing results")

exception_caught_before_debugger = Callback(
    doc="Called whenever an exception is caught, but a debugger hasn't been entered yet"
)
exception_caught_after_debugger = Callback(
    doc="Called whenever an exception is caught, and a debugger has already been run"
)

_CUSTOM_HOOKS = {}

def add_custom_hook(hook_name):
    """
    Adds an additional hook to the set of available hooks
    """
    globs = globals()
    if hook_name in _CUSTOM_HOOKS or hook_name in globs:
        raise HookAlreadyExists("Hook named {0!r} already exists!".format(hook_name))

    returned = _CUSTOM_HOOKS[hook_name] = globs[hook_name] = Callback()
    return returned

def ensure_custom_hook(hook_name):
    """
    Like :func:`.add_custom_hook`, only forgives if the hook already exists
    """
    if hook_name in _CUSTOM_HOOKS:
        return _CUSTOM_HOOKS[hook_name]
    return add_custom_hook(hook_name)

def remove_custom_hook(hook_name):
    """
    Removes a hook from the set of available hooks
    """
    _CUSTOM_HOOKS.pop(hook_name)
    globals().pop(hook_name)

def get_custom_hook_names():
    """
    Retrieves the names of all custom hooks currently installed
    """
    return list(_CUSTOM_HOOKS)

def get_all_hooks():
    for name, callback in _compat.iteritems(globals()):
        if not isinstance(callback, Callback):
            continue
        yield name, callback

def get_hook_by_name(hook_name):
    """
    Returns a hook (if exists) by its name, otherwise returns None
    """
    return _CUSTOM_HOOKS.get(hook_name, None)




import logbook
from logbook.compat import LoggingHandler
import platform
import itertools
import forge
import shutil
import tempfile
if platform.python_version() < "2.7":
    import unittest2 as unittest
else:
    import unittest

import slash
from slash.conf import config
from slash import RunnableTestFactory

_logger = logbook.Logger(__name__)

class TestCase(unittest.TestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        self._handler = LoggingHandler()
        self._handler.push_application()
        self.addCleanup(self._handler.pop_application)
        self.override_config("hooks.swallow_exceptions", False)
        self.override_config("log.console_level", 10000) # silence console in tests

    def override_config(self, path, value):
        self.addCleanup(config.assign_path, path, config.get_path(path))
        config.assign_path(path, value)

    def get_new_path(self):
        returned = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, returned)
        return returned

    _forge = None
    @property
    def forge(self):
        if self._forge is None:
            self._forge = forge.Forge()
        return self._forge
    _events = None
    @property
    def events(self):
        if self._events is None:
            self._events = self.forge.create_wildcard_mock()
        return self._events
    def tearDown(self):
        if self._forge is not None:
            self._forge.restore_all_replacements()
            self._forge.verify()
        super(TestCase, self).tearDown()

class NullFile(object):
    def write(self, s):
        pass
    def isatty(self):
        return False

class CustomException(Exception):
    @classmethod
    def do_raise(cls):
        raise cls("Custom exception")

def no_op(*args, **kwargs):
    pass

def run_tests_in_session(test_class_path_or_iterator):
    if isinstance(test_class_path_or_iterator, str):
        test_class_path_or_iterator = slash.loader.Loader().iter_paths([test_class_path_or_iterator])
    if isinstance(test_class_path_or_iterator, type) and issubclass(test_class_path_or_iterator, RunnableTestFactory):
        test_class_path_or_iterator = test_class_path_or_iterator.generate_tests()
    with slash.Session() as session:
        slash.runner.run_tests(test_class_path_or_iterator)
    for result in session.results.iter_test_results():
        for err in itertools.chain(result.get_errors(), result.get_failures(), result.get_skips()):
            _logger.debug("Unsuccessful result: {0}", err)
    return session
run_tests_in_session.__test__ = False

def run_tests_assert_success(test_class_path_or_iterator):
    session = run_tests_in_session(test_class_path_or_iterator)
    assert session.results.is_success(), "Run did not succeed"
    return session

run_tests_assert_success.__test__ = False



import sys
from .utils import TestCase
from .utils import no_op
from .utils import NullFile
from .utils.test_generator import TestGenerator
from slash.frontend import slash_run
from slash import config
from slash.frontend.main import main_entry_point
from slash._compat import StringIO
from slash import site
import os

class SlashRunTestBase(TestCase):
    def setUp(self):
        super(SlashRunTestBase, self).setUp()
        self.override_config("run.session_state_path", os.path.join(self.get_new_path(), "session_data"))


class MissingFilesTest(SlashRunTestBase):

    def test_slash_run_fails_fast_for_missing_files(self):
        result = slash_run.slash_run(["/non/existing/path"], report_stream=NullFile())
        self.assertNotEquals(result, 0, "slash run unexpectedly succeeded for a missing path")

class ArgumentParsingTest(TestCase):

    def setUp(self):
        super(ArgumentParsingTest, self).setUp()
        self.devnull = StringIO()
        self.forge.replace_with(sys, "stderr", self.devnull)

    callback_success = False
    def _test_iterator_stub(self, app, args):
        self.assertTrue(config.root.debug.enabled)
        self.assertEquals(app.args.remainder, ["test1.py", "test2.py"])
        # this must be last to make sure the stub ran successfully
        self.callback_success = True
        return ()

    def test_interspersed_positional_arguments(self):

        self.forge.replace_with(slash_run, "_get_test_iterator", self._test_iterator_stub)
        self.forge.replace_with(sys, "argv", "/path/to/slash run -vv test1.py -x test2.py --pdb".split())
        with self.assertRaises(SystemExit) as caught:
            main_entry_point()
        self.assertTrue(self.callback_success)
        if isinstance(caught.exception, int):
            # python 2.6
            code = caught.exception
        else:
            code = caught.exception.code
        self.assertEquals(code, 0)

class SlashRunTest(SlashRunTestBase):

    def setUp(self):
        super(SlashRunTest, self).setUp()
        self.generator = TestGenerator()
        make_test = self.generator.generate_test
        self.forge.replace_with(site, "load", no_op)
        self.root_path = self.generator.write_test_directory(
            {
                "test_1.py" : make_test(),
                "dir_1" : {
                    "dir_2" : {
                        "test_2.py" : [make_test(), make_test()],
                    },
                    "test_3.py" : make_test(),
                    "regular_file.txt" : "some content here",
                    "other_regular_file" : "more contents",
                }
            }, self.get_new_path()
        )

    def test_slash_run_directory_success(self):
        result = self._execute_slash_run([self.root_path])
        self.assertEquals(result, 0, "slash run did not return 0 on success")

    def test_slash_run_default_directory(self):
        self.override_config("run.default_sources", [self.root_path])
        self.assertEquals(self._execute_slash_run([]), 0)

    def test_slash_run_success_if_skips(self):
        self.generator.make_test_skip(self.generator.get_expected_test_ids()[1])
        result = self._execute_slash_run([self.root_path])
        self.assertEquals(result, 0, "skips cause nonzero return value")

    def test_slash_rerun(self):
        failing_test_id = self.generator.get_expected_test_ids()[1]
        self.generator.make_test_fail(failing_test_id)
        result = self._execute_slash_run([self.root_path])
        self.assertNotEquals(result, 0)

        self.generator.assert_all_run()

        self.generator.forget_run_test_ids()
        self.generator.make_test_fail(failing_test_id)
        result = self._execute_slash_rerun()
        self.assertNotEquals(result, 0)

        self.assertEquals(self.generator.get_test_ids_run(), [failing_test_id])

        self.generator.forget_run_test_ids()
        result = self._execute_slash_rerun()
        self.assertEquals(result, 0)

        self.assertEquals(self.generator.get_test_ids_run(), [failing_test_id])

        # tearDown will attempt to check assert_all_run...
        self.generator.reset()

    def test_slash_rerun_nonexistent_state_file(self):
        self.override_config("run.session_state_path", "/nonexisting_file")
        self.assertNotEquals(self._execute_slash_rerun(), 0)
        self.generator.reset() # don't complain over not run tests

    def test_slash_rerun_directory_state_file(self):
        self.override_config("run.session_state_path", self.get_new_path())
        self.assertNotEquals(self._execute_slash_rerun(), 0)
        self.generator.reset() # don't complain over not run tests

    def test_slash_rerun_skips(self):
        skipping_test_id = self.generator.get_expected_test_ids()[1]
        self.generator.make_test_skip(skipping_test_id)
        result = self._execute_slash_run([self.root_path])
        self.assertEquals(result, 0)
        self.generator.assert_all_run()

        self.generator.forget_run_test_ids()
        result = self._execute_slash_rerun()
        self.assertEquals(result, 0)
        self.assertEquals(self.generator.get_test_ids_run(), [])
        self.generator.reset() # tearDown will assert all tests were run

    def test_slash_run_directory_failure(self):
        self._test_slash_run_directory_unsuccessful(self.generator.make_test_fail)

    def test_slash_run_directory_error(self):
        self._test_slash_run_directory_unsuccessful(self.generator.make_test_raise_exception)

    def _test_slash_run_directory_unsuccessful(self, fault):
        expected = self.generator.get_expected_test_ids()
        fault(expected[2])
        result = self._execute_slash_run([self.root_path])
        self.assertNotEquals(result, 0, "slash run unexpectedly returned 0 for failure")

    def test_slash_run_specific_file(self):
        for path in [
                "test_1.py",
                "dir_1/dir_2/test_2.py",
                "dir_1/test_3.py"
        ]:
            result = self._execute_slash_run([os.path.join(self.root_path, path)])
            self.assertEquals(result, 0, "slash run did not return successfully for {0}".format(path))

    def _execute_slash_run(self, argv):
        return slash_run.slash_run(argv, report_stream=NullFile())

    def _execute_slash_rerun(self):
        return slash_run.slash_rerun([], report_stream=NullFile())

    def tearDown(self):
        self.generator.assert_all_run()
        super(SlashRunTest, self).tearDown()



import sys
import logbook
import slash
from slash import log
from slash._compat import cStringIO
from .utils import TestCase
from .utils.test_generator import TestGenerator
from slash.frontend import slash_run

class SlashRunTest(TestCase):
    def setUp(self):
        super(SlashRunTest, self).setUp()
        self.generator = TestGenerator()
        (self.success_test, self.fail_test,
         self.error_test, self.skip_test, self.warn_test) = self.generator.generate_tests(5)
        self.generator.make_test_skip(self.skip_test)
        self.generator.make_test_fail(self.fail_test)
        self.generator.make_test_raise_exception(self.error_test)
        self.generator.make_test_warn(self.warn_test)

        self.root_path = self.generator.write_test_directory(
            {
                "success_test.py" : self.success_test,
                "fail_test.py" : self.fail_test,
                "error_test.py" : self.error_test,
                "skip_test.py" : self.skip_test,
                "warn_test.py" : self.warn_test,
            }, self.get_new_path()
        )
        self.report_stream = cStringIO()
        self.separator = "-" * 80
        slash.hooks.session_start.register(self._silence_console_logger, "silence_console")
        self.addCleanup(slash.hooks.session_start.unregister_by_identifier, "silence_console")

    def _silence_console_logger(self):
        slash.context.session.logging.console_handler.level = 1000000

    def _get_output_part(self, part):
        self.report_stream.seek(0)
        return self.report_stream.getvalue().split(self.separator)[part]
    def get_live_part(self):
        return self._get_output_part(0)
    def get_exceptions_part(self):
        return self._get_output_part(1)
    def get_summary_part(self):
        return self._get_output_part(2)
    def _slash_run(self):
        return slash_run.slash_run([self.root_path], report_stream=self.report_stream)
    def test_live_concise(self):
        self.override_config("log.console_level", logbook.WARNING)
        self._slash_run()
        result = self.get_live_part()
        self.assertEqual(set(result) - set("\n"), set("EFS."))
        self.assertEqual(len(result), 6)

    def test_live_verbosity_info(self):
        self._test_live_verbosity(logbook.INFO)

    def test_live_verbosity_notice(self):
        self._test_live_verbosity(logbook.NOTICE)

    def _test_live_verbosity(self, level):
        self.override_config("log.console_level", level)
        self._slash_run()
        output = self.get_live_part()
        self.assertNotIn("Traceback", output)
        self._assert_reported(output, "error_test", "error")
        self._assert_reported(output, "fail_test", "fail")
        self._assert_reported(output, "skip_test", "skip")
        self._assert_reported(output, "success_test", "ok")

    def _assert_reported(self, output, test_name, additional_text):
        test = getattr(self, test_name)
        expected_output = "{0}.py:{1}.{2} ...\n{3}".format(
            test_name, test._factory_class_name, 0, additional_text)
        if expected_output not in output:
            self.fail("Expected output {0!r} not found. Output was:\n{1}".format(expected_output, output))

    def test_summary(self):
        self._slash_run()
        output = self.get_summary_part()
        headers, values, empty_line, warning_header, warning = output.strip().splitlines()
        self.assertEqual(values.split(), ['2', '1', '1', '1'])
        self.assertEqual("", empty_line) # warn
        self.assertIn("Warnings:", warning_header) # warn
        self.assertIn("This is a warning", warning) # warn
    def test_exceptions_summary_verbose(self):
        self.override_config("log.console_level", logbook.INFO)
        self._slash_run()
        output = self.get_exceptions_part()
        self.assertEqual(output.count('>'), 3) # skip, error and fail
        self.assertEqual(output.count('Traceback'), 2) # error and fail
        self.assertIn("OSError: Sample exception", output) # error
        self.assertIn("TestFailed: Test failed", output) # fail
    def test_exceptions_summary_concise(self):
        self.override_config("log.console_level", logbook.WARNING)
        self._slash_run()
        output = self.get_exceptions_part()
        self.assertEqual(output.count('>'), 3) # skip, error and fail
        self.assertEqual(output.count('Traceback'), 0) # not tracebacks!
        self.assertIn("Sample exception", output) # error
        self.assertIn("Test failed", output) # fail



import itertools
from ._compat import iteritems

def iterate(**parameters_and_values):
    """
    For each PARAM=OPTIONS passed, will cause the
    decorated test method to be actually run several times
    upon execution, one for each possible parameter
    """
    def _decorator(func):
        specs = _get_or_create_parameter_specs(func)
        for parameter_name, options in iteritems(parameters_and_values):
            specs.setdefault(parameter_name, []).extend(options)
        return func
    return _decorator

def _get_or_create_parameter_specs(func):
    returned = _get_parameter_specs(func)
    if returned is None:
        returned = func.__slash_parameters__ = {}
    return returned

def _get_parameter_specs(func):
    return getattr(func, "__slash_parameters__", None)

def iterate_kwargs_options(func):
    specs = _get_parameter_specs(func)
    if specs is None:
        yield {}
        return
    specs = list(iteritems(specs))
    names = [first for first, second in specs]
    values = [second for first, second in specs]
    for combination in itertools.product(*values): # pylint: disable=W0142
        yield dict(zip(names, combination))



class RunnableTest(object):
    """
    This class is meant to serve as a base class to any test that can
    actually be executed by the Slash runner.
    """
    __slash__ = None
    def run(self):
        """
        This method is meant to be overriden by derived classes to actually
        perform the test logic
        """
        raise NotImplementedError() # pragma: no cover

    def get_address_in_module(self):
        """
        Returns the 'address' of this test inside the module in which it resides.

        This is used in order to later refer to that test in its FQN or when rerunning it.
        """
        return None

    def __repr__(self):
        return repr(self.__slash__)



from ..interface import PluginInterface
from ...conf import config
from ...ctx import session
import requests

class Plugin(PluginInterface):
    def get_name(self):
        return "notifications"

    def activate(self):
        config.extend({
            "plugins": {
                "notifications" : {
                    "prowl_api_key" : None,
                    "nma_api_key" : None,
            }}})

    def deactivate(self):
        config["plugins"].pop("notifications")

    def session_end(self):
        self._notify("Session Ended", "Session {0} ended".format(session.id))

    def _notify(self, title, message):
        for url, api_key in [
                ("https://prowl.weks.net/publicapi/add", config.root.plugins.notifications.prowl_api_key),
                ("https://www.notifymyandroid.com/publicapi/notify", config.root.plugins.notifications.nma_api_key),
                ]:
            if api_key is None:
                continue
            resp = requests.post(url, {"apikey": api_key, "application": "slash", "event": title, "description": message})
            resp.raise_for_status()



import logbook
import itertools
from contextlib import contextmanager
from .formatter import Formatter
from .color_string import ColorString
from .hooks_context_manager import HooksContextManager
from .._compat import string_types
from ..conf import config
from ..ctx import context

def _is_concise():
    return config.root.log.console_level > logbook.NOTICE

@contextmanager
def report_context(report_stream):
    live_reporter_type = ConciseLiveReporter if _is_concise() else VerboseLiveReporter
    with live_reporter_type(report_stream):
        yield
    if _is_concise():
        report_stream.write("\n")
    SummaryReporter(report_stream).report_session(context.session)

def _format_unsuccessful(formatter, result):
    console_level = config.root.log.console_level
    with formatter.indented():
        for exc_obj in itertools.chain(result.get_failures(), result.get_errors(), result.get_skips()):
            if isinstance(exc_obj, string_types) or console_level > logbook.NOTICE:
                formatter.writeln(exc_obj)
            elif console_level > logbook.INFO:
                formatter.writeln("({0.exception!r})".format(exc_obj))
            else:
                formatter.writeln(exc_obj.exception_text)

_YELLOW = ColorString.get_formatter("yellow")
_RED = ColorString.get_formatter("red")
_GREEN = ColorString.get_formatter("green")

_REPORT_COLUMNS = [
    ("Successful", "get_num_successful", _GREEN),
    ("Failures", "get_num_failures", _RED),
    ("Errors", "get_num_errors", _RED),
    ("Skipped", "get_num_skipped", _YELLOW),
    ]

class SummaryReporter(object):
    def __init__(self, stream):
        super(SummaryReporter, self).__init__()
        self._formatter = Formatter(stream)
    def report_session(self, session):
        self._describe_unsuccessful(session)
        self._describe_summary(session)
        self._describe_warnings(session)
    def _describe_unsuccessful(self, session):
        self._formatter.write_separator()
        for result in session.results.iter_all_results():
            if result.is_success():
                continue
            self._formatter.write("> ")
            if result.is_skip():
                self._formatter.write(_YELLOW("SKIP"))
            elif result.is_error():
                self._formatter.write(_RED("ERROR"))
            elif result.is_failure():
                self._formatter.write(_RED("FAILURE"))
            else:
                raise NotImplementedError() # pragma: no cover
            if result.test_metadata is None:
                self._formatter.write(" (Outside tests)")
            else:
                self._formatter.write(" ", result.test_metadata)
            _format_unsuccessful(self._formatter, result)

    def _describe_summary(self, session):
        self._formatter.write_separator()

        summary = dict(
            (column_title, getattr(session.results, method_name)())
            for column_title, method_name, _ in _REPORT_COLUMNS
        )


        for is_header in (True, False):
            for column_title, _, colorizer in _REPORT_COLUMNS:
                column_width = len(column_title) + 2

                count = summary[column_title]
                if is_header:
                    s = column_title.ljust(column_width)
                else:
                    s = str(count)

                if count:
                    orig = s
                    s = colorizer(s)
                    column_width += len(s) - len(orig)
                self._formatter.write(s.ljust(column_width))
            self._formatter.writeln()

    def _describe_warnings(self, session):
        if session.warnings:
            self._formatter.writeln()
            self._formatter.writeln("Warnings:")
            for record in session.warnings:
                self._formatter.writeln(record)

class BaseLiveReporter(HooksContextManager):
    def __init__(self, report_stream):
        super(BaseLiveReporter, self).__init__()
        self._formatter = Formatter(report_stream)

class VerboseLiveReporter(BaseLiveReporter):
    def on_test_start(self):
        self._formatter.write("> ", context.test)
        self._formatter.writeln(" ...")
    def on_test_success(self):
        self._formatter.writeln("ok")
    def on_test_error(self):
        self._formatter.writeln("error")
    def on_test_failure(self):
        self._formatter.writeln("fail")
    def on_test_skip(self, reason): # pylint: disable=W0613
        self._formatter.writeln("skip")

class ConciseLiveReporter(BaseLiveReporter):
    def on_test_success(self):
        self._formatter.write(".")
    def on_test_error(self):
        self._formatter.write("E")
    def on_test_failure(self):
        self._formatter.write("F")
    def on_test_skip(self, reason): # pylint: disable=W0613
        self._formatter.write("S")
    def on_result_summary(self):
        self._formatter.writeln("")



import logbook
from confetti import Config
from .utils.conf_utils import Doc, Cmdline

__all__ = ["config"]

config = Config({
    "debug" : {
        "debug_skips" : False // Doc("Enter pdb also for SkipTest exceptions"),
        "enabled" : False // Doc("Enter pdb on failures and errors") // Cmdline(on="--pdb"),
    },
    "log" : {
        "console_level" : logbook.WARNING // Doc("console verbosity (can be repeated)") // Cmdline(decrease="-v", increase="-q"),
        "root" : None // Doc("Root directory for logs") // Cmdline(arg="-l", metavar="DIR"),
        "subpath" : "{context.session.id}/{context.test_id}/log" // Doc("Path to write logs to under the root"),
        "session_subpath" : "session.log",
        "silence_loggers": [] // Doc("Logger names to silence"),
        "format": None // Doc("Format of the log line, as passed on to logbook. None will use the default format"),
        "localtime": False // Doc("Use local time for logging. If False, will use UTC"),
    },
    "run" : {
        "default_sources": [] // Doc("Default tests to run assuming no other sources are given to the runner"),
        "local_customization_file_path": "./.slashrc",
        "stop_on_error" : False // Doc("Stop execution when a test doesn't succeed") // Cmdline(on="-x"),
        "session_state_path": "~/.slash/last_session" // Doc("Where to keep last session serialized data"),
        "user_customization_file_path": "~/.slash/slashrc",
    },
    "sentry" : {
        "dsn" : None // Doc("Possible DSN for a sentry service to log swallowed exceptions. "
                            "See http://getsentry.com for details"),
    },
    "hooks" : {
        "swallow_exceptions" : False // Doc("If set, exceptions inside hooks will not be re-raised"),
    },
    "plugins" : {
        "search_paths" : [] // Doc("List of paths in which to search for plugin modules"),
    },
})



from . import context
from .conf import config
from .utils.path import ensure_containing_directory
from .warnings import WarnHandler
from contextlib import contextmanager
import logbook  # pylint: disable=F0401
from logbook.more import ColorizedStderrHandler  # pylint: disable=F0401
from ._compat import ExitStack
import os

class SessionLogging(object):
    """
    A context creator for logging within a session and its tests
    """
    def __init__(self, session):
        super(SessionLogging, self).__init__()
        self.warnings_handler = WarnHandler(session.warnings)
        self.console_handler = ColorizedStderrHandler(bubble=True, level=config.root.log.console_level)
        self._set_formatting(self.console_handler)

    def get_test_logging_context(self):
        return self._get_file_logging_context(config.root.log.subpath)

    def get_session_logging_context(self):
        return self._get_file_logging_context(config.root.log.session_subpath)

    @contextmanager
    def _get_file_logging_context(self, filename_template):
        with ExitStack() as stack:
            stack.enter_context(self._get_file_log_handler(filename_template))
            stack.enter_context(self.console_handler)
            stack.enter_context(self.warnings_handler)
            stack.enter_context(self._process_test_record())
            stack.enter_context(self._get_silenced_logs_context())
            yield

    def _get_silenced_logs_context(self):
        if not config.root.log.silence_loggers:
            return ExitStack()
        return SilencedLoggersHandler(config.root.log.silence_loggers)

    def _get_file_log_handler(self, subpath):
        root_path = config.root.log.root
        if root_path is None:
            handler = logbook.NullHandler(bubble=False)
        else:
            log_path = os.path.join(root_path, subpath.format(context=context))
            ensure_containing_directory(log_path)
            handler = logbook.FileHandler(log_path, bubble=False)
            self._set_formatting(handler)
        return handler

    def _set_formatting(self, handler):
        if config.root.log.localtime:
            logbook.set_datetime_format("local")
        fmt = config.root.log.format
        if fmt is not None:
            handler.format_string = fmt

    def _add_current_test(self, record):
        record.extra['source'] = context.test_id or context.session.id

    def _process_test_record(self):
        return logbook.Processor(self._add_current_test)

class SilencedLoggersHandler(logbook.Handler):
    def __init__(self, silence_logger_names):
        super(SilencedLoggersHandler, self).__init__(bubble=False)
        self._silenced_names = set(silence_logger_names)
    def should_handle(self, record):
        return record.channel in self._silenced_names



from .utils import TestCase
from slash.utils.peekable_iterator import PeekableIterator

class PeekableIteratorTest(TestCase):
    def test_full_iteration(self):
        objects = [object() for i in range(10)]
        it = PeekableIterator(objects)
        for i, x in enumerate(it):
            self.assertIs(x, objects[i])
            for peek_num in range(3): # no matter how many times we peek, we get the same result
                if i == len(objects) - 1:
                    self.assertFalse(it.has_next())
                    with self.assertRaises(StopIteration):
                        it.peek()
                    self.assertIsNone(it.peek_or_none())
                else:
                    self.assertTrue(it.has_next())
                    self.assertIs(it.peek(), objects[i+1])
                    self.assertIs(it.peek_or_none(), objects[i+1])



"""
Test the usage of FQNs in Slash, to uniquely identify tests being run.
"""
import itertools
import os

import slash
from slash.utils.fqn import TestFQN as FQN
from slash.utils.fqn import TestPQN as PQN
from slash.utils.fqn import ModuleTestAddress, InvalidQN
from .utils import TestCase
from .utils import run_tests_assert_success

# needed for nose
FQN.__test__ = PQN.__test__ = False

class FQNTest(TestCase):

    def test_invalid_fqns(self):
        for fqn in [
                "test.py:Factory-1",
                "test.py::Factory",
                "test.py:Factory(1, 2, 3, d=2)",
                "test.py:Factory(1, 2, 3, d=2)()",
                "test.py:Factory.method(1)",
                "test.py:Factory.method(x=f(1, 2, 3))",
                #incompelte FQNs (which are legal PQNs otherwise)
                "test.py:Factory",
                ]:
            try:
                FQN.from_string(fqn)
            except InvalidQN:
                pass
            else:
                self.fail("FQN {0!r} did not fail!".format(fqn))

    
    def test_fqn_equality(self):
        f = FQN("/tmp/test.py", ModuleTestAddress("Factory", "method_name", method_kwargs={"a": 1}, before_kwargs={"b": 2}, after_kwargs={"c": 3}))
        for not_equal in [
                FQN.from_string("/tmp/test.py:Factory2.method_name"),
                FQN.from_string("/tmp/test.py:Factory.method_name"),
                FQN.from_string("/tmp/test.py:Factory.method_name(c=3)"),
                FQN.from_string("/tmp/test.py:Factory(x=1)(y=2).method_name(c=3)"),
                ]:
            self.assertTrue(not_equal != f)
            self.assertTrue(f != not_equal)
            self.assertFalse(f == not_equal)
            self.assertFalse(not_equal == f)
        for equal in [
                FQN.from_string("/tmp/test.py:Factory(b=2)(c=3).method_name(a=1)")
        ]:
            self.assertTrue(f == equal)
            self.assertTrue(equal == f)
            self.assertFalse(f != equal)
            self.assertFalse(equal != f)

class FactoryTestAddressParsingTest(TestCase):

    def test_parameters(self):
        self.assert_parsing(
            "Factory(a=1)(b=2, c=3).method(d=4, e=5)",
            factory_name="Factory",
            method_name="method",
            before_kwargs={"a": 1},
            method_kwargs={"d": 4, "e": 5},
            after_kwargs={"b": 2, "c": 3})

    def test_parameters_strings(self):
        self.assert_parsing(
            "Factory(a='string, with, comma')().method(b='another, string,')",
            factory_name="Factory",
            before_kwargs={"a": 'string, with, comma'},
            method_name="method",
            after_kwargs={},
            method_kwargs={"b": 'another, string,'})

    def test_fqn_insensitive_parameter_ordering(self):
        parse = ModuleTestAddress.from_string
        self.assertEquals(parse("F(a=1, b=2)(c=3, d=4).meth(e=5, f=6)"),
                          parse("F(b=2,a=1)(d=4,c=3).meth(f=6,e=5)"))

    def assert_parsing(self, expr, **fields):
        result = ModuleTestAddress.from_string(expr)
        for field_name, value in fields.items():
            actual_value = getattr(result, field_name)
            if actual_value != value:
                self.fail("Field {0!r} in result differs (expected: {1!r}, got {2!r})".format(
                    field_name, value, actual_value))

class PQNTest(TestCase):

    def test_is_simple_path(self):
        self.assertTrue(PQN.from_string("/tmp/testme.py").is_simple_path())
        self.assertFalse(PQN.from_string("/tmp/testme.py:bla").is_simple_path())

    def test_pyc_files_original_exists(self):
        "Filenames ending with .pyc should be normalized to .py"
        original = os.path.join(self.get_new_path(), "testme.py")
        with open(original, "w"):
            pass
        f = PQN(original+"c")
        self.assertEquals(f.path, original)

    def test_pyc_files_original_missing(self):
        "When the original python file is missing and the filename ends with .pyc, it should not be fixed"
        original = os.path.join(self.get_new_path(), "testme.pyc")
        f = PQN(original)
        self.assertEquals(f.path, original)


    def test_matching_filename(self):
        for pattern in ["/tmp/testme.py:Blap", "/tmp/testme.py", "/tmp/testme.py:Blap.method"]:
            self._assert_matches(pattern, "/tmp/testme.py:Blap.method")
            self._assert_matches(pattern, "/tmp/testme.py:Blap(a=1)(b=2).method(c=3)")
            self._assert_matches(pattern, "/tmp/testme.py:Blap.method(c=3)")
            self._assert_not_matches(pattern, "/tmp/testme2.py:F.method")
            self._assert_not_matches(pattern, "/tmp/testme2.py:F(a=1)(b=2).method(c=3)")

    def test_matching_factory(self):
        self._assert_matches("/tmp/testme.py:Factory", "/tmp/testme.py:Factory.method(a=1)")
        self._assert_matches("/tmp/testme.py:Factory", "/tmp/testme.py:Factory.method")
        self._assert_not_matches("/tmp/testme.py:Factory", "/tmp/testme.py:Factory2.method")
        self._assert_not_matches("/tmp/testme.py:Factory", "/tmp/testme2.py:Factory.method")

    def test_matching_methods(self):
        self._assert_matches("/tmp/testme.py:Factory.method", "/tmp/testme.py:Factory.method")
        self._assert_not_matches("/tmp/testme.py:Factory.method", "/tmp/testme2.py:Factory.method")
        self._assert_not_matches("/tmp/testme.py:Factory.method", "/tmp/testme.py:Factory.method2")
        self._assert_not_matches("/tmp/testme.py:Factory.method", "/tmp/testme.py:Factory2.method")
        self._assert_not_matches("/tmp/testme.py:Factory", "/tmp/testme2.py:Factory.method")

    def test_matching_parameters(self):
        for before, after, params in [
                ("(a=1)", "()", "()"),
                ("()", "(d=4)", "()"),
                ("()", "(d=4)", "(f=6)"),
                ("()", "(d=4)", "(f=6, e=5)"),
                ("(a=1, b=2)", "(d=4)", "(f=6, e=5)"),
                ("(a=1, b=2)", "(d=4, c=3)", "(f=6, e=5)"),
                ]:
            pqn = "/tmp/testme.py:Factory{0}{1}.method{2}".format(before, after, params)
            self._assert_matches(pqn, "/tmp/testme.py:Factory(a=1, b=2)(c=3, d=4).method(e=5, f=6)")
            self._assert_not_matches(pqn, "/tmp/testme.py:Factory2(a=1, b=2)(c=3, d=4).method(e=5, f=6)")
            self._assert_not_matches(pqn, "/tmp/testme.py:Factory(a=1, b=2)(c=3, d=4).method2(e=5, f=6)")
            self._assert_not_matches(pqn, "/tmp/testme.py:Factory(a=2, b=2)(c=3, d=5).method(e=5, f=7)")

    def test_invalid_pqns(self):
        for pqn in [
                "test.py:Factory-1",
                "test.py::Factory",
                "test.py:Factory(1, 2, 3, d=2)",
                "test.py:Factory(1, 2, 3, d=2)()",
                "test.py:Factory.method(1)",
                "test.py:Factory.method(x=f(1, 2, 3))",
                ]:
            try:
                PQN.from_string(pqn)
            except InvalidQN:
                pass
            else:
                self.fail("Pattern {0!r} did not fail!".format(pqn))

    def _assert_not_matches(self, p, f):
        self._assert_matches(p, f, positive=False)

    def _assert_matches(self, p, f, positive=True):
        pqn = PQN.from_string(p)
        fqn = FQN.from_string(f)

        if pqn.matches(fqn) != positive:
            msg = "does not match" if positive else "unexpectedly matches"
            self.fail("PQN {0!r} {1} FQN {2!r}".format(pqn, msg, fqn))

    def test_repr(self):
        for pattern in ["/tmp/blap:bla", "/tmp/a.py", "kjlkjdfd"]:
            self.assertEquals(repr(PQN.from_string(pattern)), pattern)



import uuid
from contextlib import contextmanager
from .. import ctx, hooks, log
from ..exception_handling import handling_exceptions
from ..interfaces import Activatable
from .result import SessionResults
from ..utils.id_space import IDSpace
from ..warnings import SessionWarnings

class Session(Activatable):
    def __init__(self):
        super(Session, self).__init__()
        self.id = "{0}:0".format(uuid.uuid1())
        self.id_space = IDSpace(self.id)
        self._complete = False
        self._context = None
        self.warnings = SessionWarnings()
        self.logging = log.SessionLogging(self)
        #: an aggregate result summing all test results and the global result
        self.results = SessionResults()

    def activate(self):
        assert self._context is None
        self._context = _session_context(self)
        with handling_exceptions():
            self._context.__enter__()

    def deactivate(self):
        self.results.global_result.mark_finished()
        with handling_exceptions():
            self._context.__exit__(None, None, None)

    def mark_complete(self):
        self._complete = True

    def is_complete(self):
        return self._complete

@contextmanager
def _session_context(session):
    assert ctx.context.session is None
    ctx.push_context()
    ctx.context.session = session
    try:
        with session.logging.get_session_logging_context():
            hooks.session_start()
            try:
                yield
            finally:
                hooks.session_end()
    finally:
        ctx.pop_context()




import ast
import os
from .._compat import string_types, iteritems

class TestPQN(object):
    """
    Represents a partially qualified name of a test.

    This can be thought of as a pattern to be matched against real FQNs
    """

    def __init__(self, path, address_in_module=None):
        super(TestPQN, self).__init__()
        assert address_in_module is None or isinstance(address_in_module, ModuleTestAddress), \
            "address_in_module must be a ModuleTestAddress object"
        self.path = path
        self.address_in_module = address_in_module
        self.path, self.abspath = self._normalize_paths(path)

        self.fqn = self.path

        if self.address_in_module is not None:
            self.fqn += ":{0}".format(self.address_in_module.to_string())

    def _normalize_paths(self, path):
        if path.endswith(".pyc"):
            non_pyc = path[:-1]
            if os.path.isfile(non_pyc):
                path = non_pyc
        abspath = os.path.abspath(path)
        path = os.path.relpath(abspath, ".")

        if path.startswith(".."):
            path = abspath

        return path, abspath

    @classmethod
    def from_string(cls, s):
        path = s
        address_in_module = None
        if ":" in path:
            path, address_in_module = path.split(":", 1)
            address_in_module = ModuleTestAddress.from_string(address_in_module)

        return cls(
            path=path, address_in_module=address_in_module)

    def is_simple_path(self):
        """
        Returns True if this pattern simply leads to a single path on the filesystem, rather than a partial match inside the file/directory
        """
        return self.address_in_module is None

    def matches(self, other):
        if not isinstance(other, TestFQN):
            return False

        if self.path != other.path:
            return False

        if self.address_in_module is not None:
            for address_field_name in ("factory_name", "method_name"):
                address_field_value = getattr(self.address_in_module, address_field_name)
                if address_field_value is not None and address_field_value != getattr(other.address_in_module, address_field_name):
                    return False

            for keywords_field_name in ("before_kwargs", "after_kwargs", "method_kwargs"):
                pattern_kwargs = getattr(self.address_in_module, keywords_field_name)
                other_kwargs = getattr(other.address_in_module, keywords_field_name)
                if not all(other_kwargs.get(arg_name, _NOTHING) == arg_value
                           for arg_name, arg_value in iteritems(pattern_kwargs)):
                    return False

        return True

    def __repr__(self):
        return self.fqn

_NOTHING = object()

class TestFQN(TestPQN):
    """
    Represents a fully-qualified name of a test being run
    """
    def __init__(self, path, address_in_module):
        if address_in_module is None or not address_in_module.is_complete():
            raise InvalidQN()
        super(TestFQN, self).__init__(path, address_in_module)

    def __eq__(self, other):
        if type(other) is not TestFQN and not isinstance(other, string_types):
            return NotImplemented

        return str(self) == str(other)

    def __ne__(self, other):
        return not (self == other)

class ModuleTestAddress(object):
    """
    Represents the 'address' of a test inside the factory that generated it
    """

    def __init__(self, factory_name, method_name=None, before_kwargs=None, after_kwargs=None, method_kwargs=None):
        super(ModuleTestAddress, self).__init__()
        self.factory_name = factory_name
        self.method_name = method_name
        self.before_kwargs = before_kwargs or {}
        self.after_kwargs = after_kwargs or {}
        self.method_kwargs = method_kwargs or {}

    def is_complete(self):
        return self.factory_name is not None and self.method_name is not None

    @classmethod
    def from_string(cls, s):
        try:
            node = ast.parse(s)
        except SyntaxError:
            raise InvalidQN()
        if len(node.body) != 1 or not isinstance(node.body[0], ast.Expr):
            raise InvalidQN()
        node = node.body[0].value
        return ModuleTestAddress(**_parse_test_module_address_ast(node))

    def to_string(self):
        returned = self.factory_name
        if self.before_kwargs or self.after_kwargs:
            returned += "{0}{1}".format(
                self._get_call_string(self.before_kwargs),
                self._get_call_string(self.after_kwargs))
        if self.method_name is not None:
            returned += ".{0}".format(self.method_name)
            if self.method_kwargs:
                returned += self._get_call_string(self.method_kwargs)
        return returned

    def _get_call_string(self, call):
        return "({0})".format(", ".join("{0}={1!r}".format(k, v) for k, v in iteritems(call)))

    def __eq__(self, other):
        if type(other) is not ModuleTestAddress:
            return NotImplemented

        return all(getattr(self, x) == getattr(other, x) for x in
                   ("factory_name", "method_name", "before_kwargs", "after_kwargs", "method_kwargs"))

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return self.to_string()

def _parse_test_module_address_ast(node):
    if isinstance(node, ast.Call):
        if isinstance(node.func, ast.Call):
            # factory call
            return _parse_factory_class_ast(node)
        elif isinstance(node.func, ast.Attribute):
            # method call
            returned = _parse_method_ast(node.func)
            returned["method_kwargs"] = _parse_call(node)
        else:
            raise InvalidQN()
    elif isinstance(node, ast.Attribute):
        # method getting
        returned = _parse_factory_class_ast(node.value)
        returned["method_name"] = node.attr
    elif isinstance(node, ast.Name):
        returned = {"factory_name":  node.id, "method_name": None}
    else:
        raise InvalidQN()
    return returned

def _parse_method_ast(node):
    returned = _parse_factory_class_ast(node.value)
    returned["method_name"] = node.attr
    return returned

def _parse_factory_class_ast(node):
    returned = {}
    if isinstance(node, ast.Call):
        if not isinstance(node.func, ast.Call) or not isinstance(node.func.func, ast.Name):
            raise InvalidQN()

        returned["after_kwargs"] = _parse_call(node)
        returned["before_kwargs"] = _parse_call(node.func)
        returned["factory_name"] = node.func.func.id
    elif isinstance(node, ast.Name):
        returned["factory_name"] = node.id
    else:
        raise InvalidQN()
    return returned

def _parse_call(node):
    if node.args or node.starargs or node.kwargs:
        raise InvalidQN()
    try:
        return dict((kw.arg, ast.literal_eval(kw.value)) for kw in node.keywords)
    except ValueError:
        raise InvalidQN()


class InvalidQN(ValueError):
    def __init__(self):
        super(InvalidQN, self).__init__("Invalid FQN given")



#! /usr/bin/python
from __future__ import print_function
import subprocess
import sys
import os

_PYPY = hasattr(sys, "pypy_version_info")

def _execute(cmd):
    if 0 != subprocess.call(cmd, shell=True):
        sys.exit(-1)

if __name__ == '__main__':
    print("Running from", os.path.abspath("."))
    deps = [
        "coverage",
        "coveralls",
        "nose",
        "pyforge",
        "lxml", # for XSD validations
    ]
    if not _PYPY:
        deps.append("pylint>=1.0.0")
    if sys.version_info < (2, 7):
        deps.append("unittest2")

    _execute("pip install --use-mirrors {0}".format(" ".join(repr(dep) for dep in deps)))

    _execute("python setup.py develop")

    if not _PYPY:
        _execute("pylint --rcfile=.pylintrc setup.py")
        _execute("pylint --rcfile=.pylintrc slash")
    _execute("coverage run $(which nosetests) -w tests")



from __future__ import absolute_import
import sys
import traceback

def get_traceback_string(exc_info=None):
    if exc_info is None:
        exc_info = sys.exc_info()
    exc_type, exc_value, exc_tb = exc_info
    return "".join(traceback.format_exception(exc_type, exc_value, exc_tb))



from .utils import TestCase
from slash._compat import OrderedDict
from slash.core.result import Result
from slash.core.result import SessionResults

class SessionResultTest(TestCase):
    def setUp(self):
        super(SessionResultTest, self).setUp()
        self.results = [
            Result() for _ in range(10)
            ]
        # one result with both errors and failures
        try:
            1 / 0
        except:
            self.results[1].add_error()
            self.results[1].add_failure()
            # one result with failure
            self.results[2].add_failure()
            # one result with error
            self.results[3].add_error()
            self.results[5].add_error()

        # one result will skip
        self.results[4].add_skip("Reason")

        # and one result will skip with error
        self.results[5].add_skip("Reason")
        num_finished = 7

        for result in self.results[:num_finished]:
            result.mark_finished()
        self.result = SessionResults()
        for index, r in enumerate(self.results):
            self.result._results_dict[index] = r

    def test_counts(self):
        self.assertEquals(self.result.get_num_results(), len(self.results))
        self.assertEquals(self.result.get_num_successful(), 2)
        # errors take precedence over failures
        self.assertEquals(self.result.get_num_errors(), 3)
        self.assertEquals(self.result.get_num_skipped(), 2)
        self.assertEquals(self.result.get_num_failures(), 1)



import sys
from .utils.fqn import TestFQN, ModuleTestAddress

class Metadata(object):

    def __init__(self, test, factory=None, factory_index=0):
        super(Metadata, self).__init__()
        self.id = None
        self.factory = factory
        self.factory_index = factory_index

        factory_class = factory if factory is not None else type(test)
        address_in_module = test.get_address_in_module()
        if address_in_module is None:
            # the test does not know how to report itself...
            address_in_module = ModuleTestAddress(factory_name=factory_class.__name__, method_name=factory_index)

        self.fqn = TestFQN(
            path=sys.modules[factory_class.__module__].__file__,
            address_in_module=address_in_module,
            )

    def __repr__(self):
        return repr(self.fqn)

def ensure_test_metadata(test):
    returned = getattr(test, "__slash__", None)
    if returned is None:
        returned = test.__slash__ = Metadata(test)
    return returned



import os
import json
from .path import ensure_directory
from ..conf import config

def save_session_state(state):
    path = _get_path()
    ensure_directory(os.path.dirname(path))
    with open(path, "w") as f:
        json.dump(state, f)

def get_last_session_state():
    with open(_get_path(), "r") as f:
        return json.load(f)

def _get_path():
    return os.path.expanduser(config.root.run.session_state_path)



from .utils import TestCase
from .utils import CustomException
from slash.utils.callback import Callback

class CallbackTestBase(TestCase):
    def setUp(self):
        super(CallbackTestBase, self).setUp()
        self.hook = Callback(["arg_value"])
        self.arg = object()

class CallbackTest(CallbackTestBase):
    def test_cannot_call_with_positionals(self):
        with self.assertRaises(TypeError):
            self.hook(1)
    def test_callback(self):
        callback1 = self.forge.create_wildcard_function_stub()
        callback2 = self.forge.create_wildcard_function_stub()
        self.hook.register(callback1)
        self.hook.register(callback2)
        # expect
        callback1(arg_value=self.arg)
        callback2(arg_value=self.arg)
        self.forge.replay()
        self.hook(arg_value=self.arg)
    def test_callback_registering_via_decorator(self):
        handler = self.forge.create_wildcard_function_stub()
        @self.hook.register
        def proxy(arg_value):
            handler(arg_value=arg_value)

        self.assertIsNotNone(proxy)
        #expect
        handler(arg_value=self.arg)
        self.forge.replay()
        self.hook(arg_value=self.arg)

class HookExceptionsTest(CallbackTestBase):
    def setUp(self):
        super(HookExceptionsTest, self).setUp()
        self.callbacks = [self.forge.create_wildcard_function_stub() for _ in range(5)]
        for index, callback in enumerate(self.callbacks):
            self.hook.register(callback)
            expected_call = callback(arg_value=self.arg)
            if index % 3 == 0:
                expected_call.and_raise(CustomException(index))
    def test_swallow_exceptions(self):
        self.override_config("hooks.swallow_exceptions", True)
        self.forge.replay()
        self.hook(arg_value=self.arg)
    def test_dont_swallow_exceptions(self):
        self.override_config("hooks.swallow_exceptions", False)
        self.forge.replay()
        with self.assertRaises(CustomException) as caught:
            self.hook(arg_value=self.arg)
        self.assertEquals(caught.exception.args[0], 0, "First exception was not the one propagated from hook!")



from contextlib import contextmanager
from .utils.debug import debug_if_needed
from . import hooks as trigger_hook
from .conf import config
import functools
import logbook
try:
    import raven # pylint: disable=F0401
except ImportError:
    raven = None
import sys

_logger = logbook.Logger(__name__)

def trigger_hooks_before_debugger(_):
    trigger_hook.exception_caught_before_debugger()
def trigger_hooks_after_debugger(_):
    trigger_hook.exception_caught_after_debugger()

_EXCEPTION_HANDLERS = [
    trigger_hooks_before_debugger,
    debug_if_needed,
    trigger_hooks_after_debugger,
    ]

@contextmanager
def handling_exceptions(**kwargs):
    try:
        yield
    except:
        handle_exception(sys.exc_info(), **kwargs)
        raise

def handle_exception(exc_info, context=None):
    """
    Call any handlers or debugging code before propagating an exception onwards.

    This makes sure that the exception can be handled as close as possible to its originating point.

    .. note:: this *DOES NOT* take care of adding the error to the session or test results!
    """
    msg = "Handling exception"
    if context is not None:
        msg += " (Context: {0})"
    _logger.debug(msg, context, exc_info=exc_info)
    if not is_exception_handled(exc_info[1]):
        for handler in _EXCEPTION_HANDLERS:
            handler(exc_info)
    mark_exception_handled(exc_info[1])

def mark_exception_handled(e):
    mark_exception(e, "handled", True)

def is_exception_handled(e):
    """
    Checks if the exception ``e`` already passed through the exception handling logic
    """
    return bool(get_exception_mark(e, "handled", False))

_NO_DEFAULT = object()

def is_exception_marked(e, name):
    return get_exception_mark(e, name, _NO_DEFAULT) is not _NO_DEFAULT

def mark_exception(e, name, value):
    """
    Associates a mark with a given value to the exception ``e``
    """
    _ensure_exception_marks(e)[name] = value

def get_exception_mark(e, name, default=None):
    """
    Given an exception and a label name, get the value associated with that mark label.
    If the label does not exist on the specified exception, ``default`` is returned.
    """
    return _ensure_exception_marks(e).get(name, default)

def _ensure_exception_marks(e):
    returned = getattr(e, "__slash_exc_marks__", None)
    if returned is None:
        returned = e.__slash_exc_marks__ = {}
    return returned

@contextmanager
def get_exception_swallowing_context(report_to_sentry=True):
    """
    Returns a context under which all exceptions are swallowed (ignored)
    """
    try:
        yield
    except:
        if not get_exception_mark(sys.exc_info()[1], "swallow", True):
            raise
        if report_to_sentry:
            capture_sentry_exception()
        _logger.debug("Ignoring exception", exc_info=sys.exc_info())

def noswallow(exception):
    """
    Marks an exception to prevent swallowing by :func:`slash.exception_handling.get_exception_swallowing_context`,
    and returns it
    """
    mark_exception(exception, "swallow", False)
    return exception

def disable_exception_swallowing(func_or_exception):
    """
    Marks an exception to prevent swallowing. Can also be used as a decorator around a function to mark all escaped
    exceptions
    """
    if isinstance(func_or_exception, BaseException):
        return noswallow(func_or_exception)
    @functools.wraps(func_or_exception)
    def func(*args, **kwargs):
        try:
            return func_or_exception(*args, **kwargs)
        except BaseException as e:
            disable_exception_swallowing(e)
            raise
    return func

def capture_sentry_exception():
    client = get_sentry_client()
    if client is not None:
        client.captureException()

def get_sentry_client():
    if raven is not None and config.root.sentry.dsn:
        return raven.Client(config.root.sentry.dsn)
    return None



import slash
from .utils import TestCase

class FailuresAndErrorsTest(TestCase):

    def test_adding_errors_global(self):
        self._test_adding_errors_failures_strings(
            slash.add_error,
            lambda result: result.get_errors(),
            False,
        )

    def test_adding_errors_in_test(self):
        self._test_adding_errors_failures_strings(
            slash.add_error,
            lambda result: result.get_errors(),
            True,
        )

    def test_adding_failures_global(self):
        self._test_adding_errors_failures_strings(
            slash.add_failure,
            lambda result: result.get_failures(),
            False,
        )

    def test_adding_failures_in_test(self):
        self._test_adding_errors_failures_strings(
            slash.add_failure,
            lambda result: result.get_failures(),
            True
        )

    def _test_adding_errors_failures_strings(self, adder, getter, in_specific_test):
        obj = SampleObject()
        class Test(slash.Test):
            def test(self):
                adder("msg1")
                adder("msg2")
                adder(obj)


        with slash.Session() as session:
            if in_specific_test:
                slash.run_tests(slash.loader.Loader().iter_test_factory(Test))
                [result] = session.results.iter_test_results()
            else:
                list(Test.generate_tests())[0].run()
                result = session.results.global_result

        self.assertFalse(result.is_success())

        [error1, error2, error3] = getter(result)
        self.assertEquals(error1.message, "msg1")
        self.assertEquals(error2.message, "msg2")
        self.assertEquals(error3.message, None)
        self.assertIs(error3.arg, obj)

class SampleObject(object):
    pass



import itertools
import functools
from ..ctx import context
from .._compat import itervalues
from .._compat import OrderedDict
from ..utils.error_object import Error

class Result(object):

    def __init__(self, test_metadata=None):
        super(Result, self).__init__()
        self.test_metadata = test_metadata
        self._errors = []
        self._failures = []
        self._skips = []
        self._finished = False
        self._interrupted = False

    def is_error(self):
        return bool(self._errors)

    def is_failure(self):
        return bool(self._failures)

    def is_just_failure(self):
        """Indicates this is a pure failure, without errors involved"""
        return self.is_failure() and not self.is_error()

    def is_skip(self):
        return bool(self._skips)

    def is_success(self, allow_skips=False):
        returned = not self._errors and not self._failures and not self._interrupted
        if not allow_skips:
            returned &= not self._skips
        return returned

    def is_success_finished(self):
        return self.is_success() and self.is_finished()

    def is_finished(self):
        return self._finished

    def mark_finished(self):
        self._finished = True

    def mark_interrupted(self):
        self._interrupted = True

    def is_interrupted(self):
        return self._interrupted

    def add_error(self, msg=None):
        self._add_error(self._errors, msg)

    def add_failure(self, msg=None):
        self._add_error(self._failures, msg)

    def _add_error(self, error_list, msg):
        error_list.append(Error(msg))

    def add_skip(self, reason):
        self._skips.append(reason)

    def get_errors(self):
        return self._errors

    def get_failures(self):
        return self._failures

    def get_skips(self):
        return self._skips

    def __repr__(self):
        return "< Result ({0})>".format(
            ", ".join(
                attr
                for attr in ("success", "error", "failure", "skip", "finished", "interrupted")
                if getattr(self, "is_{0}".format(attr))()
            )
        )


class GlobalResult(Result):
    pass


class SessionResults(object):

    def __init__(self):
        super(SessionResults, self).__init__()
        self.global_result = GlobalResult()
        self._results_dict = OrderedDict()
        self._iterator = functools.partial(itervalues, self._results_dict)

    @property
    def current(self):
        test_id = context.test_id
        if test_id is None:
            return self.global_result
        return self._results_dict[test_id]

    def __iter__(self):
        return self._iterator()

    def is_success(self, allow_skips=False):
        return self.global_result.is_success() and \
            all(result.is_finished() and result.is_success(allow_skips=allow_skips)
                for result in self._iterator())

    def get_num_results(self):
        return len(self._results_dict)

    def get_num_successful(self):
        return self._count(Result.is_success_finished, include_global=False)

    def get_num_errors(self):
        return self._count(Result.is_error)

    def get_num_failures(self):
        return self._count(Result.is_just_failure)

    def get_num_skipped(self):
        return self._count(Result.is_skip)

    def _count(self, pred, include_global=True):
        returned = 0
        iterator = self.iter_all_results(
        ) if include_global else self.iter_test_results()
        for result in iterator:
            if pred(result):
                returned += 1
        return returned

    def iter_test_results(self):
        return iter(self)

    def iter_all_results(self):
        return itertools.chain(self.iter_test_results(), [self.global_result])

    def create_result(self, test):
        assert test.__slash__.id not in self._results_dict
        returned = Result(test.__slash__)
        self._results_dict[test.__slash__.id] = returned
        return returned

    def get_result(self, test):
        if test.__slash__ is None:
            raise LookupError("Could not find result for {0}".format(test))
        return self._results_dict[test.__slash__.id]

    def __getitem__(self, test):
        return self.get_result(test)
